/****************************************************************************************************\
 *  GPGPU exercise:                                                                                 *
 *  compute Mandelbrot fractal images on the GPU                                                    *
 *  (all your programming should take place in this file; the others are support infrastructure)    *
 *                                                                                                  *
 *  tasks:                                                                                          *
 *  - port the given reference implementation to CUDA                                               *
 *  - run both the serial and the parallel implementations for the initial parameter values         *
 *    (left, right, top, bottom, maxIter as given in the source code)                               *
 *  - inspect and compare the two results                                                           *
 *    (ue_dumpImageToFile() and ue_countDifferingArrayElements() in utility_environment.h might be  *
 *    useful)                                                                                       *
 *                                                                                                  *
 *  If all goes well, the GPU implementation should be fast enough that you can interactively       *
 *  explore the fractal. Use the arrow keys to scroll, and PgUp/PgDown to zoom.                     *
 *                                                                                                  *
 *  hints:                                                                                          *
 *  - don't be alarmed if the CPU and GPU results are not 100% bit-wise identical                   *
 *  - the hardest part is converting from blockIdx and threadIdx to x and y coordinates             *
 *                                                                                                  *
 *  additional tinkering (optional):                                                                *
 *  - vary block size of the CUDA kernel call (powers of 2) and observe effect on (GPU) run time    *
 *    (ue_startTimer() and ue_stopTimer() in utility_environment.h might be useful)                 *
\****************************************************************************************************/

// include some convenience functions (names begin with "ue_" to distinguish them from CUDA)
#include "utility_environment.h"

// OpenGL utilities
#include "opengl_includes.h"
#include "CudaTimer.h"

////////////////////////////////////////////////////////////////////////////////
// GLUT, which we use for displaying images, makes use of callback functions. //
// The following global variables are the simplest way of passing parameters  //
// between the various callback functions.                                    //
////////////////////////////////////////////////////////////////////////////////

bool g_useCuda = false;
// boundaries of fractal parameter space (these effectively control location and zoom)
float g_left = -2.0f;
float g_right = 0.5f;
float g_top = 1.25f;
float g_bottom = -1.25f;

int g_imgWidth = 0;
int g_imgHeight = 0;

int *h_gpuResult = 0;                 // host buffer for copy of GPU results
int *d_gpuResult = 0;                 // target buffer in device memory, for GPU results
int *h_cpuResult = 0;                 // target buffer in host memory, for reference results

CPUBitmap *g_bitmap = 0;

int g_maxIter = 255;

/******************************\
 *  reference implementation  *
\******************************/

// iterate one point through fractal formula
int iterate_cpu(float cx, float cy, int maxIter)
{
    float x = 0.0f, y = 0.0f;           // tracking the complex number x + y*i
    float x2 = x*x, y2 = y*y;           // these hold squared values
    int iter = 0;

    do
	{
        iter++;
	    y = 2.0f*x*y + cy;              // complex square, so we have "z_next := z*z + c"
	    x = x2 - y2 + cx;
	    y2 = y*y;                       // compute squares of new values
	    x2 = x*x;
    } while ((iter <= maxIter) && (x2 + y2 < 4.0f));

	if (iter > maxIter)
	{
		iter = 0;
	}

    return iter;
}

// for all image pixels, compute parameter values and perform fractal iterations
void fractal_cpu(int *h_ptr, int width, int height, float left, float right, float top, float bottom, int maxIter)
{
    for (int y = 0; y < height; y++) 
	{
	    float fy = top + (bottom - top)*(float(y)/float(height - 1));      // map from pixel coordinates to parameter space
	    for (int x = 0; x < width; x++) 
		{
	        float fx = left + (right - left)*(float(x)/float(width - 1)); // map from pixel coordinates to parameter space
		    h_ptr[y*width + x] = iterate_cpu(fx, fy, maxIter);
	    }
	}
}


/******************************************************\
 *  GPGPU implementation (to be done by the STUDENT)  *
\******************************************************/

// for all image pixels, compute parameter values and perform fractal iterations
__global__ void fractal_gpu(int *d_ptr, int width, int height, float left, float right, float top, float bottom, int maxIter)
{
	
}


// compute and draw image
void computeFractal(int imgWidth, int imgHeight, int *d_gpuResult, int *h_cpuResult, int *h_gpuResult)
{
	CudaTimer timer;
	timer.start();
	
	if ( g_useCuda )
	{
		//////////////////////////////////////////////////////////////////////////////////////////////
		// call CUDA kernel to run on the GPU, copy results back to host, and measure required time //
		//////////////////////////////////////////////////////////////////////////////////////////////
		printf("Running Cuda code ...\n");
		// set threads layout;  we utilize the fact that img_size = k*32 (see above)
		// consequence: blocks.x * threads.x = img_size
		// -> we don't need to check in the kernel whether the thread is outside the image!
		dim3 cudaBlockSize(16,16);  // image is subdivided into rectangular tiles for parallelism - this variable controls tile size
		dim3 cudaGridSize(imgWidth/cudaBlockSize.x, imgHeight/cudaBlockSize.y);
		fractal_gpu<<<cudaGridSize, cudaBlockSize>>>(d_gpuResult, imgWidth, imgHeight, g_left, g_right, g_top, g_bottom, g_maxIter);
		cudaDeviceSynchronize();  // wait until GPU is finished
		handleCUDAerror(__LINE__);

		// copy GPU result back to host memory
		cudaMemcpy(h_gpuResult, d_gpuResult, imgWidth*imgHeight*sizeof(int), cudaMemcpyDeviceToHost);
		handleCUDAerror(__LINE__);
	}
	else // CPU implementation
	{
		printf("Running CPU code ...\n");
		fractal_cpu(h_gpuResult, imgWidth, imgHeight, g_left, g_right, g_top, g_bottom, g_maxIter);
	}
	printf( "Calculation finished after %5.2f milliseconds.\n", timer.elapsed() );
}

void cleanUp()
{
	// free memory buffers
    cudaFree(d_gpuResult);
    cudaFreeHost(h_cpuResult);
	handleCUDAerror(__LINE__);

    cudaFreeHost(h_gpuResult);
}

void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
	{
	case ' ':
		g_useCuda = !g_useCuda;
		break;
	case 27:              // keycode for the Escape key
		cleanUp();
		exit(0);
		break;
	}

	glutPostRedisplay();
}

void specialKeyboard(int key, int x, int y)
{
	const float panSpeed = 0.03f;    // both always positive, and larger values mean larger steps
	const float zoomSpeed = 0.1f;

	float width = g_right - g_left;  // temporary values for scrolling and zooming
	float height = g_top - g_bottom;
	float center_x = 0.5f*(g_left + g_right);
	float center_y = 0.5f*(g_top + g_bottom);

	switch (key) 
	{
	case GLUT_KEY_LEFT:              // cursor left key: scroll left
		g_left -= panSpeed*width;
		g_right -= panSpeed*width;
		break;
	case GLUT_KEY_RIGHT:             // cursor right key: scroll right
		g_left += panSpeed*width;
		g_right += panSpeed*width;
		break;
	case GLUT_KEY_UP:                // cursor up key: scroll up
		g_top -= panSpeed*height;
		g_bottom -= panSpeed*height;
		break;
	case GLUT_KEY_DOWN:              // cursor down key: scroll down
		g_top += panSpeed*height;
		g_bottom += panSpeed*height;
		break;
	case GLUT_KEY_PAGE_UP:           // page up key: zoom out from display center
		g_left = center_x - (1.0f + zoomSpeed) * 0.5f * width;
		g_right = center_x + (1.0f + zoomSpeed) * 0.5f * width;
		g_top = center_y + (1.0f + zoomSpeed) * 0.5f * height;
		g_bottom = center_y - (1.0f + zoomSpeed) * 0.5f * height;
		break;
	case GLUT_KEY_PAGE_DOWN:         // page down key: zoom into display center
		g_left = center_x - 0.5f * width / (1.0f + zoomSpeed);
		g_right = center_x + 0.5f * width / (1.0f + zoomSpeed);
		g_top = center_y + 0.5f * height / (1.0f + zoomSpeed);
		g_bottom = center_y - 0.5f * height / (1.0f + zoomSpeed);
		break;
	}
	glutPostRedisplay();
}

// calls computation, pixel conversion, and OpenGL draw when GLUT refreshes the display
void display( ) 
{
	computeFractal(g_imgWidth, g_imgHeight,	d_gpuResult, h_cpuResult, h_gpuResult);
	ue_convertFractalIterations2Bitmap(h_gpuResult, g_imgWidth, g_imgHeight, 0, g_maxIter, g_bitmap);
	g_bitmap->draw();
	glFlush();
}


/************************
 *  Setup and Test Rig  *
\************************/

// reads image size from the first command line parameter
// ensures that the value is divisible by 32 and has a reasonable default
int fetchImageExtent(int argc, char **argv)
{
    int imgSize = 0;            // width and height of square 2D image 

	if ( argc > 1 )             // interpret first command line argument as image size
	{
		imgSize = atoi( argv[1] );
	}
	if (imgSize == 0)
	{
		imgSize = 512;          // default value
	}

	return (imgSize/32) * 32;   // round up to a multiple of 32
}


int main(int argc, char **argv)
{
	g_imgHeight = g_imgWidth = fetchImageExtent(argc, argv);  // read image size from command line

	///////////////////////////////////////////////////////////////////////////
	// initialize GLUT, open display window, and register callback functions //
	///////////////////////////////////////////////////////////////////////////
	glutInit( &argc, argv );
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );
	glutInitWindowSize( g_imgWidth, g_imgHeight );
	glutCreateWindow( "Fractal - arrow keys to pan, PgUp/PgDown to zoom - press ESC to close" );
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(specialKeyboard);
	glutDisplayFunc(display);

	///////////
	// setup //
	///////////
	ue_checkGPU();                 // verify that a CUDA device is available
	
	// allocate buffer(s) in main memory and on the GPU
    cudaMalloc(&d_gpuResult, g_imgWidth*g_imgHeight*sizeof(int));
    cudaMallocHost(&h_cpuResult, g_imgWidth*g_imgHeight*sizeof(int));
    cudaMallocHost(&h_gpuResult, g_imgWidth*g_imgHeight*sizeof(int));
	handleCUDAerror(__LINE__);     // in principle, every single CUDA call should be followed by an
	                               // error check. we skipped some to reduce visual clutter. you can
								   // always add the checks to hunt down bugs.
	g_bitmap = new CPUBitmap(g_imgWidth, g_imgHeight);

	glutMainLoop();

	return 0;
}
